/*
    ldapdiff
    Copyright (C) 2000-2008 Thomas.Reith@rhoen.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <lber.h>
#include <ldap.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ldapdiff.h"

static char ldifescape[128] = {
 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1,

 0, 0, 0, 0, 0, 0, 0, 0,
 1, 1, 1, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0,

 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 1, 0, 0, 0,

 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 1
};

#define LDIFNEEDFLTESCAPE(c) ((c) & 0x80 || ldifescape[(unsigned)(c)])

static size_t ldif_bv2escaped_filter_value_len(char *val,size_t val_len)
{
 size_t i,l;

 if(val == NULL || val_len == 0){
  ldiflog(LOG0,"fatal error: file %s line %d",__FILE__,__LINE__);
  exit(EXITLDERROR);
 }

 for(l = 0,i = 0;i < val_len;l++,i++){
  char c = val[i];
  
  if(LDIFNEEDFLTESCAPE(c)){
   l += 2;
  }
 }
 return l;
}

void ldif_bv2escaped_filter_value(char *inval,size_t inval_len,char **poutval)
{
 size_t  i,l,o;
 char   *out;

 if(inval == NULL || inval_len == 0 || poutval == NULL){
  ldiflog(LOG0,"fatal error: file %s line %d",__FILE__,__LINE__);
  exit(EXITLDERROR);
 }

 l = ldif_bv2escaped_filter_value_len(inval,inval_len);
 out = LDALLOC(l + 1,sizeof(char));
 *poutval = out;

 if(l == inval_len){
  memcpy(out,inval,inval_len);
  return;
 }

 o = 0;
 for(i = 0;i < inval_len;i++){
  char c = inval[i];
  
  if(LDIFNEEDFLTESCAPE(c)){
   out[o++] = '\\';
   out[o++] = "0123456789ABCDEF"[0x0f & (c>>4)];
   out[o++] = "0123456789ABCDEF"[0x0f & c];
  } 
  else{
   out[o++] = c;
  }
 }
 out[o] = '\0';

 return;
}

int ldifgetfilter(struct s_ldifentry *sldifentry,char **filter)
{
 struct s_ldifentry *psldifentry;
 struct s_ldifval   *psldifval;

 char *cfilter = ldifgetpconf(CONFFILTER);
 char *cgroup  = ldifgetpconf(CONFGROUP);

 psldifentry = sldifentry;
 while(psldifentry != NULL){
  if(strcasecmp(cfilter,psldifentry->var) == 0){
   char   *cval;
   size_t  filter_len;

   psldifval = psldifentry->vallist;

   ldif_bv2escaped_filter_value(psldifval->val,psldifval->val_len,&cval);
   filter_len = strlen("(&()(=))") + 
                strlen(cgroup)     + 
                strlen(cfilter)    +
                strlen(cval)       +
                1;
   *filter = LDALLOC(filter_len,sizeof(char));
   /*
   this statement crashes on fc8 08.12.2007 for large cval > 30k
   sprintf(*filter,"(&(%s)(%s=%s))",cgroup,cfilter,cval);
   */
   strcat(*filter,"(&(");
   strcat(*filter,cgroup);
   strcat(*filter,")(");
   strcat(*filter,cfilter);
   strcat(*filter,"=");
   strcat(*filter,cval);
   strcat(*filter,"))");

   if(psldifval->val_len != strlen(cval)){
ldiflog(LOG0,"hossa [%s]",cval);
   } 

   if(psldifval->next != NULL){
    ldiflog(LOG0,"error: filter attribute [%s] not single valued",cfilter);
    exit(EXITLDERROR);
   }

   free(cval);
   return 0;
  }
  psldifentry = psldifentry->next;
 }
 return -1;
}

void ldiffiltercheck(struct s_schema *sschema)
{
 char *cfilter = ldifgetpconf(CONFFILTER);
    
 if(ldifschemaequal(sschema,cfilter) == 0){
  ldiflog(LOG0,"error: filter attribute [%s] cannot be used, because",cfilter);
  ldiflog(LOG0,"error: the ldap schema does not define EQUALITY method",cfilter);
  exit(EXITLDERROR);
 }
}
